/* CViewColumn.cp - This View is a columnar table of sub-views.

	by David Phillip Oster 1997
 */
#include "CViewColumn.h"

CViewColumn::CViewColumn(LStream *inStream) : LView(inStream){
}

/* GetPaneCount - return the number of sub panes in this column
 */
SInt32 CViewColumn::GetPaneCount()const{
	return mSubPanes.GetCount();
}

/* GetSubPane - return a reference to the pane at the 1's based index
	if you override this you'll probably also have to override all other places in this
		file that explicitly reference mSubPanes
 */
LPane*	CViewColumn::GetSubPane(SInt32 index){
	return const_cast<LPane*>(GetConstSubPane(index));
}

/* GetConstSubPane - be picky about constness.
 */
const LPane* CViewColumn::GetConstSubPane(SInt32 index)const{
	LPane*	pIth;

	if(mSubPanes.FetchItemAt(index, pIth)){
		return pIth;
	}
	return NULL;
}

/* CreateSubPane - 
	installs the pane verticaly after its predecessors, and visually moves 
		those that follow it down to make room.
 */
LPane* CViewColumn::CreateSubPane(SInt32 index, ResIDT inPPobID, Boolean inRefresh){
	SetDefaultAttachable(NULL);

	void			*valp = NULL;
	LPane			*pane;
	LPane			*pIth;
	SInt32			height;
	SDimension16	dim;
	SDimension32	mySize;
	LCommander*		saveCommander;

	saveCommander = LCommander::GetDefaultCommander();
	try{
		valp = UReanimator::ReadObjects('PPob', inPPobID);
	}catch(LException &err){
		LCommander::SetDefaultCommander(saveCommander);
		throw;
	}
	LCommander::SetDefaultCommander(saveCommander);
	pane = reinterpret_cast<LPane*>(valp);
	pane = dynamic_cast<LPane*>(pane);

	Assert_(pane);
	ThrowIfNil_(pane);

	pane->PutInside(this, false);

	// make an attempt to order the subPanes array the way the caller asked.
	if(NULL == (pIth = GetSubPane(index))  || pane != pIth){
		mSubPanes.Remove(pane);
		mSubPanes.InsertItemsAt(1, index, pane);
	}

	// visually place after previous pane
	height = 0;
	if(NULL != (pIth = GetSubPane(index-1))){
		SPoint32		ithLoc;
		SPoint32		myLoc;
		SPoint32		myScroll;
		SDimension16	ithSiz;
		GetFrameLocation(myLoc);
		GetScrollPosition(myScroll);
		pIth->GetFrameLocation(ithLoc);
		pIth->GetFrameSize(ithSiz);
		height = ithLoc.v + myScroll.v - myLoc.v + ithSiz.height;
	}
	pane->GetFrameSize(dim);
	GetImageSize(mySize);
	if(mySize.height < height + dim.height){
		ResizeImageBy(0, height + dim.height - mySize.height, inRefresh);
	}
	pane->PlaceInSuperImageAt(0, height, inRefresh);

	// place panes after this visually after this
	if(index != mSubPanes.GetCount()){
		SInt32			i;

		for(i = index+1;i <= mSubPanes.GetCount();i++){
			if(NULL != (pIth = GetSubPane(i))){
				pIth->MoveBy(0,  dim.height, inRefresh);
			}
		}
	}

	pane->FinishCreate();
	pane->Show();

	return pane;
}

/* RemoveSubPane - 
	slide those that follow it up by the size of it.
 */
void CViewColumn::RemoveSubPane(LPane *inSub){
	SInt32			index;
	SInt32			i;
	LPane*			pIth;
	SDimension16	dim;

	if(LArray::index_Bad != (index = mSubPanes.FetchIndexOf(inSub))){
		inSub->GetFrameSize(dim);
		mSubPanes.RemoveItemsAt(1, index);
		for(i = index;i <= mSubPanes.GetCount();i++){
			if(NULL != (pIth = GetSubPane(i))){
				pIth->MoveBy(0, - dim.height, Refresh_Yes);
			}
		}
		ResizeImageBy(0, -dim.height, Refresh_Yes);
	}
}

/* GetSuperViewColumn - if this pane has a super view which is a CViewColumn, return that
	CViewColumn, and its direct descendent in our geneology. Used by ScrollSubPaneIntoFrame
	return NULL if not found.
	[static]
 */
void CViewColumn::GetSuperViewColumn(LPane* inMe, CViewColumn* &outSuper, LPane* &outSub){
	LView	*super;
	LPane	*lastSuper;

	for(lastSuper = inMe; NULL != (super = inMe->GetSuperView()) ; lastSuper = inMe, inMe = super){
		if(NULL != (outSuper = dynamic_cast<CViewColumn*>(super))){
			outSub = lastSuper;
			return;
		}
	}
	outSuper = NULL;
	outSub = NULL;
}


/* ScrollSubPaneIntoFrame - call this from, say an editfield, to scroll its owning
	view element into visibility, if it is not.
	[static]
 */
void CViewColumn::ScrollSubPaneIntoFrame(LPane *inSub){
	CViewColumn	*super;
	LPane		*sub;
	Rect		subFrame;
	Rect		frame;
	SInt32		vertScroll;

	GetSuperViewColumn(inSub, super, sub);
	if(NULL != super && 
		super->CalcPortFrameRect(frame) &&
		sub->CalcPortFrameRect(subFrame)){

		vertScroll = 0;			// Find vertical scroll amount

		if(frame.bottom < subFrame.bottom){	// Cell is below bottom of frame
			vertScroll = subFrame.bottom - frame.bottom;
		}else if(subFrame.top < frame.top){	// Cell is above top of frame
			vertScroll = subFrame.top - frame.top;
		}
		super->UpdatePort();
		super->ApplyForeAndBackColors();
		super->ScrollPinnedImageBy(0, vertScroll, Refresh_Yes);
	}
}
/* GetValue - the value of a CViewColumn the integer index of the subView that is
	at the top edge of the CViewColumn. 1's based. return 0 if not found.
 */
SInt32 CViewColumn::GetValue() const{
	SInt32			y;
	SPoint32		pos;
	SDimension16	size;

	GetScrollPosition(pos);
	y = pos.v;
	LArrayIterator	iter(mSubPanes);
	LPane*			pIth;
	while(iter.Next(&pIth)){
		if(y <= 0){
			return iter.GetCurrentIndex();
		}
		pIth->GetFrameSize(size);
		y -= size.height;
	}
	return 0;
}

/* SetValue - setting the value of a CViewColumn makes it scroll so that the iTh
	subView is at the top edge of the CViewColumn. 1's based.
 */
void CViewColumn::SetValue(SInt32 inValue){
	LPane*			pIth;
	if(NULL != (pIth = GetSubPane(inValue))){
		SPoint32		ithLoc;
		SPoint32		loc;

		GetFrameLocation(loc);
		pIth->GetFrameLocation(ithLoc);
		ScrollImageBy(0, ithLoc.v - loc.v, Refresh_Yes);
	}
}

