/* CViewDivider.cp - this view controls two sub panes. It mouse tracks the space
	between them, as a splitter, and resizes the two sub panes based on the motion.
	It tracks the cursor.
	When you resize it, the subpanes resize proportionately.

	Assumes the sub panes are listed in the PPob in the correct order (i.e., the
	subpane closest to (0,0) first), and that their bound/unbound bits are set
	correctly (i.e., unbound in the dimension of the divider's travel, bound on 
	the other dimension.

	 you can now shrink a horizontal View to a width of 0, then grow it again, and it
		remembers the proportions of the subpanes. (similarly for vertical)

	 you now specify a minimum in the .r file, and whether it is allowed to
		shrink a pane with a minimum to zero.

	 minimums are enforced for mouse clicking and dragging on the divider.


	Current directions:	(look for a "not yet" comment in the code)

	 AdaptToSuperFrameSize needs work for panes with minimums.
	 MoveDividerBySelf and AdaptToSuperFrameSize need work to support the CanVanish booleans.


	Future directions:

	 should the per subPane data move into an array of records so we can extend this, and the .r
		to support an arbitrary number of subPanes?

	 We need a way of asserting and enforcing constraints on panes (CViewColumn
		can be configured to want to be a multiple of its element size high)

	This could be expanded to a class that took an arbitrary number of
	subviews, and dynamicly used the gaps between them as movable dividers.
	(if you change this, you shoulld also re-test CSuperViewDivider, since
		it depends on the implementation of this.)

	by David Phillip Oster 1997
 */
#include "CViewDivider.h"
#include <URegions.h>
#include <algorithm>


CViewDivider::CViewDivider(LStream *inStream) : LView(inStream){
	inStream->ReadBlock(&mFirstID, sizeof mFirstID);
	inStream->ReadBlock(&mFirstMinSize, sizeof mFirstMinSize);
	inStream->ReadBlock(&mFirstCanVanish, sizeof mFirstCanVanish);

	inStream->ReadBlock(&mSecondID, sizeof mSecondID);
	inStream->ReadBlock(&mSecondMinSize, sizeof mSecondMinSize);
	inStream->ReadBlock(&mSecondCanVanish, sizeof mSecondCanVanish);

	mIsHorizontal = 3;	// so we can tell if we were initialized.
	mFirstRatio = 0.8f;	// arbitrary. must be [0..1]
}

CViewDivider::~CViewDivider(){
}

void CViewDivider::FinishCreateSelf(){
	inherited::FinishCreateSelf();
	RecalcIsHorizontal();
	RecalcFirstRatio();

}


/* GetFrameRect - return the image coordinates of the pane
 */
static void GetFrameRect(const LPane* pane, Rect& r){
	if(nil == pane){
		r.left = r.top = r.right = r.bottom = 0;
	}else{
		SDimension16	paneS;
		SPoint32		paneL;

		pane->GetFrameSize(paneS);
		pane->GetFrameLocation(paneL);
		r.left = paneL.h;
		r.top = paneL.v;
		r.right = r.left + paneS.width;
		r.bottom = r.top + paneS.height;
	}
}

/* RecalcIsHorizontal - set mIsHorizontal true if this is a horizontal divider.
[protected] Note: we do this once, in FinishCreateSelf
 */
void CViewDivider::RecalcIsHorizontal(){
	LPane	*first, *second;
	Rect	firstR, secondR;

	first = FindPaneByID(mFirstID);
	second = FindPaneByID(mSecondID);

	Assert_(first);
	Assert_(second);

	GetFrameRect(first, firstR);
	GetFrameRect(second, secondR);

	mIsHorizontal = (firstR.right < secondR.left);
}

/* RecalcFirstRatio - preserve the dimensions as a ratio so that if
	we are recovering from a shrink to zero, we can get the aspect ratio correct.
 */
void CViewDivider::RecalcFirstRatio(){
	LPane*			first;
	SDimension16	mySize;
	SDimension16	firstSize;
	
	GetFrameSize(mySize);
	if(nil != (first = FindPaneByID(mFirstID))){
		first->GetFrameSize(firstSize);
		if(mIsHorizontal){
			if(0 < mySize.width){
				mFirstRatio = firstSize.width / (float) mySize.width;
			}
		}else{
			if(0 < mySize.height){
				mFirstRatio = firstSize.height / (float) mySize.height;
			}
		}
	}
}


/* GetDividerRect - get the visible rect of this pane in port coordinates
 */
void CViewDivider::GetDividerRect(Rect& r){
	LPane	*first, *second;
	Rect	firstR, secondR;

	first = FindPaneByID(mFirstID);
	second = FindPaneByID(mSecondID);

	Assert_(first);
	Assert_(second);

	GetFrameRect(first, firstR);
	GetFrameRect(second, secondR);

	Assert_(3 != mIsHorizontal);

	if(mIsHorizontal){
		r.left = firstR.right;
		r.right = secondR.left;
		r.top = firstR.top;
		r.bottom = firstR.bottom;
	}else{
		r.top = firstR.bottom;
		r.bottom = secondR.top;
		r.left = firstR.left;
		r.right = firstR.right;
	}
}

void CViewDivider::DrawSelf(){
	Rect		r;
	RGBColor	saveFore;
	static RGBColor	darkRGB = { 0xBBBB, 0xBBBB, 0xBBBB };
	static RGBColor lightRGB = { 0xFFFF, 0xFFFF, 0xFFFF };
	static RGBColor lidRGB = { 0xDDDD, 0xDDDD, 0xDDDD };

	inherited::DrawSelf();
	FocusDraw();

	GetDividerRect(r);
	PortToLocalPoint(topLeft(r));
	PortToLocalPoint(botRight(r));

#if 1
	GetForeColor(&saveFore);
	RGBForeColor(&lightRGB);
	MoveTo(r.left, r.bottom-1);
	LineTo(r.left, r.top);
	LineTo(r.right-1, r.top);
	RGBForeColor(&darkRGB);
	MoveTo(r.right-1, r.top+1);
	LineTo(r.right-1, r.bottom-1);
	LineTo(r.left+1, r.bottom-1);
	RGBForeColor(&lidRGB);
	InsetRect(&r, 1, 1);
	PaintRect(&r);
	RGBForeColor(&saveFore);
#else
	PaintRect(&r);
#endif
}



/* AdjustCursorSelf - Set the cursor shape when the cursor is over a divider
 */
void CViewDivider::AdjustMouseSelf(Point inPortPt, const EventRecord& inMacEvent, RgnHandle ioRgn){
	Rect		r;
	CursHandle	cursor;

	GetDividerRect(r);
	if(PtInRect(inPortPt, &r)){
		if(nil != (cursor = GetCursor(mIsHorizontal ? curs_HorizontalDivider : curs_VerticalDivider))){
			SetCursor(*cursor);
		}
	}else{
		inherited::AdjustMouseSelf(inPortPt, inMacEvent, ioRgn);
	}
}

/* MoveDividerBy - move a horizontal divider horizontally, or a vertical divider vertically by delta pixels.
	(you can't make a sub pane smaller than zero size: the deltas are clipped.)

	we separate it out into MoveBy and MoveBySelf to support CSuperViewDivider
 */
void CViewDivider::MoveDividerBy(SInt16 dx, SInt16 dy, Boolean inRefresh){
	MoveDividerBySelf(dx, dy, inRefresh);
}

/* MoveDividerBySelf - move a horizontal divider horizontally, or a vertical divider vertically by delta pixels.
	(you can't make a sub pane smaller than zero size: the deltas are clipped.)
 */
void CViewDivider::MoveDividerBySelf(SInt16 dx, SInt16 dy, Boolean inRefresh){
	LPane	*first, *second;
	Rect	firstR, secondR;
	SInt16	deltaX, deltaY;

	if(mIsHorizontal){
		if(0 == dx){
			return;
		}
		first = FindPaneByID(mFirstID);
		second = FindPaneByID(mSecondID);

		Assert_(first);
		Assert_(second);

		GetFrameRect(first, firstR);
		GetFrameRect(second, secondR);

		deltaX = dx;
		dx = std::max(mFirstMinSize - (firstR.right - firstR.left), std::min(static_cast<int>(dx), secondR.right - secondR.left - mSecondMinSize));
		if(mFirstCanVanish || mSecondCanVanish){	// not yet
		}
		first->ResizeFrameBy(dx, 0, inRefresh);
		second->MoveBy(dx, 0, inRefresh);
		second->ResizeFrameBy(-dx, 0, inRefresh);
	}else{
		if(0 == dy){
			return;
		}
		first = FindPaneByID(mFirstID);
		second = FindPaneByID(mSecondID);

		Assert_(first);
		Assert_(second);

		GetFrameRect(first, firstR);
		GetFrameRect(second, secondR);

		deltaY = dy;
		dy = std::max(mFirstMinSize - (firstR.bottom - firstR.top) , std::min(static_cast<int>(dy), secondR.bottom - secondR.top - mSecondMinSize));
		if(mFirstCanVanish || mSecondCanVanish){	// not yet
		}
		first->ResizeFrameBy(0, dy, inRefresh);
		second->MoveBy(0, dy, inRefresh);
		second->ResizeFrameBy(0, -dy, inRefresh);
	}
	RecalcFirstRatio();
}

/* GetDragLimitSlop - compute parameters for DragGrayRgn in ClickSelf
 */
void CViewDivider::GetDragLimitSlop(Rect& outR, Rect& outLimitR, Rect& outSlopR){
	GetDividerRect(outR);
	PortToLocalPoint(topLeft(outR));
	PortToLocalPoint(botRight(outR));
	CalcLocalFrameRect(outLimitR);
	if(mIsHorizontal){
		if( ! mFirstCanVanish){
			outLimitR.left += mFirstMinSize;
		}
		if( ! mSecondCanVanish){
			outLimitR.right -= mSecondMinSize;
		}
	}else{
		if( ! mFirstCanVanish){
			outLimitR.top += mFirstMinSize;
		}
		if( ! mSecondCanVanish){
			outLimitR.bottom -= mSecondMinSize;
		}
	}
	outSlopR = outLimitR;
	::InsetRect(&outSlopR, -100, -100);
}

/* ClickSelf - 
 */
void CViewDivider::ClickSelf(const SMouseDownEvent& inMouseDown){
	// Now do the actual dragging
	Point		delta;
	Rect		r, limitRect, slopRect;
	StRegion	dragRgn;

	FocusDraw();
	GetDragLimitSlop(r, limitRect, slopRect);
	::RectRgn(dragRgn, &r);

	*(SInt32*) &delta = ::DragGrayRgn(dragRgn, inMouseDown.whereLocal,
						&limitRect, &slopRect, mIsHorizontal ? hAxisOnly : vAxisOnly, nil);

	if(mIsHorizontal){
		if( ! (0 == delta.h || (SInt16) 0x8000 == delta.h)){
			MoveDividerBy(delta.h, 0, true);
		}
	}else{
		if( ! (0 == delta.v || (SInt16) 0x8000 == delta.v)){
			MoveDividerBy(0, delta.v, true);
		}
	}
}

/* AdaptToSuperFrameSize - assumes that for a 
	horizontal divider, the sub panes stick top and bottom,
	vertical divider, the sub panes stick left and right.

	Note: when making the transition from zero width to non-zero width, arbitrarily give 
		all the space to the first pane, minus a four pixel divider.
 */
void CViewDivider::AdaptToSuperFrameSize(
	SInt32	inSurrWidthDelta,			// Change in width of SuperView
	SInt32	inSurrHeightDelta,			// Change in height of SuperView
	Boolean	inRefresh){

	inherited::AdaptToSuperFrameSize(inSurrWidthDelta, inSurrHeightDelta, inRefresh);

	LPane	*first, *second;
	Rect	firstR, secondR;
	SInt16	firstPart, secondPart;
	SDimension16	mySize;

	GetFrameSize(mySize);
	if(mIsHorizontal){
		if(0 != inSurrWidthDelta){

			first = FindPaneByID(mFirstID);
			second = FindPaneByID(mSecondID);

			Assert_(first);
			Assert_(second);

			GetFrameRect(first, firstR);
			GetFrameRect(second, secondR);
			firstPart = inSurrWidthDelta * mFirstRatio;
			secondPart = inSurrWidthDelta - firstPart;
			if(0 != mFirstMinSize){		// not yet
			}
			if(0 != mSecondMinSize){	// not yet
			}
			first->ResizeFrameBy(firstPart, 0, inRefresh);
			second->MoveBy(firstPart, 0, inRefresh);
			second->ResizeFrameBy(secondPart, 0, inRefresh);
		}
	}else {
		if(0 != inSurrHeightDelta){

			first = FindPaneByID(mFirstID);
			second = FindPaneByID(mSecondID);

			Assert_(first);
			Assert_(second);

			GetFrameRect(first, firstR);
			GetFrameRect(second, secondR);
			firstPart = inSurrHeightDelta * mFirstRatio;
			secondPart = inSurrHeightDelta - firstPart;
			if(0 != mFirstMinSize){		// not yet
			}
			if(0 != mSecondMinSize){	// not yet
			}
			first->ResizeFrameBy(0, firstPart, inRefresh);
			second->MoveBy(0, firstPart, inRefresh);
			second->ResizeFrameBy(0, secondPart, inRefresh);
		}
	}
}

